{
package eGuideDog::Browser::HTML::Navigate;

use strict;
use HTML::Element;

sub new;
sub find;
sub node_index;
sub index_node;
sub current_node;
sub restart; # move to the first valid node
sub moveto_next_node;
sub moveto_previous_node;
sub moveto_next_branch;
sub moveto_previous_branch;
sub moveto_next_valid_node;
sub moveto_next_part;
sub moveto_previous_part;
sub is_slient_branch;
sub is_sound_branch;
sub build_tabindex;

sub new {
  my ($class, $tree) = @_;
#  bless $tree, 'HTML::Element';

  my $self = {root => $tree,
	      nodes => undef, # valid nodes buffer
	      node_index => 0, # index of nodes buffer
	      parent_node => $tree,
	      current_node => $tree->content_array_ref->[0],
	      current_index => 0, # index in parent node
	     };

  bless $self, __PACKAGE__;
  return $self;
}

sub find {
  my ($self, $keyword) = @_;

  my $old_index = $self->{node_index};
  $self->{node_index}++;
  while ($self->{node_index} < $self->{nodes}
	 && ($self->index_node->as_text !~ /$keyword/i
	     || $self->index_node->tag eq 'div'
	     || $self->index_node->tag eq 'ul'
	     || $self->index_node->tag eq 'ol')) {
    $self->{node_index}++;
  }

  if ($self->{node_index} < $self->{nodes}) {
    return 1;
  } else {
    $self->{node_index} = 0;
    while ($self->{node_index} <= $old_index
	   && ($self->index_node->as_text !~ /$keyword/i
	       || $self->index_node->tag eq 'div'
	       || $self->index_node->tag eq 'ul'
	       || $self->index_node->tag eq 'ol')) {
      $self->{node_index}++;
    }

    if ($self->{node_index} <= $old_index) {
      return -1;
    } else {
      $self->{node_index} = $old_index;
      return 0;
    }
  }
}

sub node_index {
  my $self = shift;
  return $self->{node_index};
}

sub index_node {
  my $self = shift;
  return $self->{nodes}->[$self->{node_index}];
}

sub current_node {
  my $self = shift;
  return $self->{current_node};
}

# move to the first valid node
sub restart {
  my $self = shift;
  $self->{parent_node} = $self->{root};
  $self->{current_node} = $self->{root}->content_array_ref->[0];
  $self->{current_index} = 0;
  $self->{node_index} = 0;
  return $self->moveto_next_valid_node();
}

# move to the next html node
sub moveto_next_node {
  my $self = shift;

  my %clone = %{$self};

  if (ref $self->{current_node}) {
    # if there is child, return the first
    my $child = $self->{current_node}->content_array_ref->[0];
    if ($child) {
      $self->{current_index} = 0;
      $self->{parent_node} = $self->{current_node};
      return $self->{current_node} = $child;
    }
  }

 repeat_find_next_node:
  # if there is right sibling, return it
  $self->{current_index}++;
  my $right = $self->{parent_node}->content_array_ref()->[$self->{current_index}];
  if ($right) {
    return $self->{current_node} = $right;
  } else {
    $self->{current_index}--;
  }

  # if there is parent, return it
  my $grandpa = $self->{parent_node}->parent;
  if ($grandpa) {
    $self->{current_node} = $self->{parent_node};
    $self->{parent_node} = $grandpa;
    $self->{current_index} = $self->{current_node}->pindex;
    goto repeat_find_next_node;
  }

  %{$self} = %clone;
  return undef;
}

# move to the previous html node
sub moveto_previous_node {
  my $self = shift;
  my %clone = %{$self};

  # find it's left node if exists
  if ($self->{current_index} > 0) {
    $self->{current_index}--;
    my $left = $self->{parent_node}->content_array_ref->[$self->{current_index}];

    # then find the last child
  repeat_find_previous_node:
    if (ref $left) {
      my $len = @{$left->content_array_ref};
      if ($len > 0) {
	$self->{parent_node} = $left;
	$self->{current_index} = $len - 1;
	$left = $left->content_array_ref->[$len - 1];
	goto repeat_find_previous_node;
      } else {
	$self->{current_node} = $left;
      }
    } else {
      $self->{current_node} = $left;
    }
    return $self->{current_node};
  }

  # if no left, then get the parent node
  if ($self->{parent_node}->parent) {
    $self->{current_node} = $self->{parent_node};
    $self->{current_index} = $self->{parent_node}->pindex;
    $self->{parent_node} = $self->{parent_node}->parent;
    return $self->{current_node};
  }

  %{$self} = %clone;
  return undef;
}

# move to the next branch of html node
sub moveto_next_branch {
  my $self = shift;
  my %clone = %{$self};

 repeat_find_next_branch:
  # if there is right sibling, return it
  $self->{current_index}++;
  my $right = $self->{parent_node}->content_array_ref->[$self->{current_index}];
  if ($right) {
    return $self->{current_node} = $right;
  }

  # if there is parent, return it
  my $grandpa = $self->{parent_node}->parent;
  if ($grandpa) {
    $self->{current_node} = $self->{parent_node};
    $self->{parent_node} = $grandpa;
    $self->{current_index} = $self->{current_node}->pindex;
    goto repeat_find_next_branch;
  }

  %{$self} = %clone;
  return undef;
}

sub moveto_previous_branch {
  my $self = shift;
  my %clone = %{$self};

  # find its left node if exists
 repeat_find_previous_branch:
  if ($self->{current_index} > 0) {
    $self->{current_index}--;
    my $left = $self->{parent_node}->content_array_ref->[$self->{current_index}];
    $self->{current_node} = $left;
    return $self->{current_node};
  }

  # if no left, then get the parent node
  if ($self->{parent_node}->parent) {
    $self->{current_node} = $self->{parent_node};
    $self->{current_index} = $self->{parent_node}->pindex;
    $self->{parent_node} = $self->{parent_node}->parent;
    if (ref $self->{current_node}
	&& !is_silent_node($self->{current_node})) {
      return $self->{current_node};
    } else {
      goto repeat_find_previous_branch;
    }
  }

  %{$self} = %clone;
  return undef;
}

sub moveto_next_valid_node {
  my $self = shift;
  my %clone = %{$self};

  if ($self->is_sound_branch($self->{current_node})) {
    goto undef_node if (!$self->moveto_next_branch());
  } else {
    goto undef_node if (!$self->moveto_next_node());
  }

  while (1) {
    # move one big stop when occured these elements
    if ($self->is_silent_branch($self->{current_node})) {
      goto undef_node if (!$self->moveto_next_branch());
    }

    # move one small step when occured these elements
    elsif ($self->is_silent_node($self->{current_node})) {
      goto undef_node if (!$self->moveto_next_node());
    }

    # done
    else {
      return $self->{current_node};
    }
  }

 undef_node:
  %{$self} = %clone;
  return undef;
}

sub moveto_next_part {
  my $self = shift;
  my %clone = %{$self};

  while (1) {
    $self->{current_node} = $self->{nodes}->[++$self->{node_index}];
    if ($self->{current_node}) {
      if (ref $self->{current_node}
	  && $self->{current_node}->tag eq 'div'
	  && $self->{current_node}->attr('class') eq 'e-guidedog_part') {
	return $self->{current_node};
      }
    } else {
      %{$self} = %clone;
      return undef;
    }
  }
}

sub moveto_previous_part {
  my $self = shift;
  my %clone = %{$self};

  while (1) {
    if ($self->{node_index} > 0) {
      $self->{current_node} = $self->{nodes}->[--$self->{node_index}];
      if (ref $self->{current_node}
	  && $self->{current_node}->tag eq 'div'
	  && $self->{current_node}->attr('class') eq 'e-guidedog_part') {
      if (ref $self->{current_node}) {
#	debug($self->{current_node}->starttag);
      } else {
#	debug($self->{current_node});
      }
	return $self->{current_node};
      }
    } else {
      %{$self} = %clone;
      return undef;
    }
  }
}

# check whether the html node can be skiped
# return 1: yes
# return 0: no
sub is_silent_node {
  my ($self, $node) = @_;

  if (ref $node) {
    bless $node, 'HTML::Element';

    if ($node->tag eq 'head'
	|| $node->tag eq 'meta'
	|| $node->tag eq 'area'
	|| $node->tag eq 'option'
	|| $node->tag eq 'link'
	|| $node->tag eq 'body'
	|| $node->tag eq 'base'
	|| $node->tag eq 'noframes'
	|| $node->attr('disabled')
	|| $node->tag eq 'label'
	|| $node->tag eq 'fieldset'
	|| $node->tag eq 'noscript'

	# table elements
	|| $node->tag eq 'table'
	|| $node->tag eq 'caption'
	|| $node->tag eq 'thead'
	|| $node->tag eq 'tfoot'
	|| $node->tag eq 'tbody'
	|| $node->tag eq 'tr'
	|| $node->tag eq 'td'
	|| $node->tag eq 'dl' # definition list
	|| $node->tag eq 'dt' # definition term
	|| $node->tag eq 'dd' # definition data

	|| $node->tag =~ /^h.$/
	|| ($node->tag eq 'li'
	    && ($node->attr('type') ne 'e-guidedog_paragraph'))
	|| ($node->tag eq 'div'
	    && ($node->attr('class') ne 'e-guidedog_paragraph')
	    && ($node->attr('class') ne 'e-guidedog_table')
#	    && ($node->attr('class') ne 'e-guidedog_text')
	    && ($node->attr('class') ne 'e-guidedog_part'))
	|| $node->tag eq 'span'
	|| $node->tag eq 'spacer' # this is found in Yahoo, not in HTML 4.01
	|| $node->tag eq 'p'
	|| $node->tag eq 'pre'
	|| $node->tag eq 'br'
	|| $node->tag eq 'blockquote' # long quotations
	|| $node->tag eq 'q' # short quotations
	|| $node->tag eq 'sub' # sbuscripts
	|| $node->tag eq 'sup' # superscripts
	|| $node->tag eq 'pre'

	# font style, not necessary to treated here but in rule_restruct.pl
	|| $node->tag eq 'em' # EM: Indicates emphasis.
	|| $node->tag eq 'strong' # STRONG: Indicates stronger emphasis.
	|| $node->tag eq 'cite' # CITE: Contains a citation or a reference to other sources.
	|| $node->tag eq 'dfn' # DFN: Indicates that this is the defining instance of the enclosed term.
	|| $node->tag eq 'code' # CODE: Designates a fragment of computer code.
	|| $node->tag eq 'samp' # SAMP: Designates sample output from programs, scripts, etc.
	|| $node->tag eq 'kbd' # KBD: Indicates text to be entered by the user.
	|| $node->tag eq 'var' # VAR: Indicates an instance of a variable or program argument.
	|| $node->tag eq 'abbr' # ABBR: Indicates an abbreviated form (e.g., WWW, HTTP, URI, Mass., etc.).
	|| $node->tag eq 'acronym' # ACRONYM: Indicates an acronym (e.g., WAC, radar, etc.).
	|| $node->tag eq 'b'
	|| $node->tag eq 'i'
	|| $node->tag eq 'ins'
	|| $node->tag eq 'del'

	|| $node->tag eq 'form'
	|| ($node->tag eq 'img' && !$node->attr('usemap'))
	|| $node->tag eq 'input' && $node->attr('type') eq 'hidden'
	|| $node->tag eq 'center') {
      return 1;
    }
  }

  return 0;
}

# check whether the html branch can be skiped
# return 1 for yes
# return 0 for no
sub is_silent_branch {
  my ($self, $node) = @_;

  if (ref $node) {
    bless $node, 'HTML::Element';

    if ($node->tag eq 'map'
	|| $node->tag eq 'script') {
      return 1;
    }
  }

  return 0;
}

# check whether the html branch is spoke as a whole
sub is_sound_branch {
  my ($self, $node) = @_;

  if (ref $node) {
    bless $node, 'HTML::Element';

    if ($node->tag eq 'title'
	|| $node->tag eq 'select'
	|| $node->tag eq 'legend'
	|| ($node->tag eq 'div'
	    && $node->attr('class') eq 'e-guidedog_table') # table with caption or th or many cells is treated real data table. Or it is treated structure table.
#	|| $node->tag eq 'li'
	|| $node->tag eq 'object'
	|| ($node->tag eq 'li' && $node->attr('type') eq 'e-guidedog_paragraph')
	|| ($node->tag eq 'div' && $node->attr('class') eq 'e-guidedog_paragraph')
	|| ($node->tag eq 'div' && $node->attr('class') eq 'e-guidedog_text')
	|| $node->tag eq 'a') {
      return 1;
    }
  }

  return 0;
}

sub build_tabindex {
  my $self = shift;
  my %clone = %{$self};

  my @tabindex;
  my $len = 0;
  while ($self->moveto_next_valid_node()) {
    if (ref $self->{current_node} && $self->{current_node}->attr('tabindex')) {
      my $new_index = $self->{current_node}->attr('tabindex');
      if ($new_index >= 0) {
	if ($len == 0) {
	  push(@tabindex, $self->{current_node});
	} else {
	  push(@tabindex, '');
	  my $i = $len - 1;
	  while ($i >= 0 && $tabindex[$i]->attr('tabindex') > $new_index) {
	    $tabindex[$i + 1] = $tabindex[$i];
	    $i--;
	  }
	  $tabindex[$i + 1] = $self->{current_node};
	}
	$len++;
      }
    }
  }
  %{$self} = %clone;
  if ($len > 0) {
    $self->{tabindex} = \@tabindex;
  }
}
}

1;
